/*
 *
 * ====================================================================
 *
 * The Apache Software License, Version 1.1
 * JVending
 * Copyright (c) 2005 Shane Isbell.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by
 *        JVending (http://jvending.sourceforge.net/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "JVending" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact shane.isbell@gmail.com.
 *
 * 5. Products derived from this software may not be called "JVending"
 *    nor may "JVending" appear in their names without prior written
 *    permission of Shane Isbell.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 *
 */


package org.jvending.registry;

import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParser;

import java.util.Hashtable;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.io.InputStream;
import java.io.IOException;
import java.io.FileInputStream;

/**
 * The central registry for all repositories. On a load method, the class will read the configuration file and
 * dynamically instantiate the repository given within the repository-config tag.
 * Sample config file
 * <pre>
&lt;registry-config&gt;
   &lt;repositories&gt;
      &lt;repository&gt;
         &lt;repository-name&gt;adapter&lt;/repository-name&gt;
         &lt;repository-class&gt;org.jvending.sample.MyRepository&lt;/repository-class&gt;
         &lt;repository-config&gt;${basedir}/adapters.txt&lt;/repository-config&gt;
	    &lt;init-param&gt;
		    &lt;param-name&gt;add&lt;/param-name&gt;
		    &lt;param-value&gt;true&lt;/param-value&gt;
	    &lt;/init-param&gt;
        &lt;init-param&gt;
		    &lt;param-name&gt;delete&lt;/param-name&gt;
		    &lt;param-value&gt;false&lt;/param-value&gt;
	    &lt;/init-param&gt;
      &lt;/repository&gt;
    &lt;/repositories&gt;
&lt;/registry-config&gt;
     </pre>
* For example to access the PropertyRepository, merely requires:
 <pre>
    RepositoryRegistry.loadFromFile("./sample-config.xml");
    PropertyRepository repository = (PropertyRepository) RepositoryRegistry.find("adapter"); `
</pre>
 *
 * If you have loaded other repositories into the registry and need to access them through the find method, you would
 * need to cast to the appropriate type.
 *
 * @author Shane Isbell
 * @version 0.12
 * @since 0.1
 */

    //TODO: Only lock methods on writes
    //TODO: Option to load configuration files from a URL

public final class RepositoryRegistry {

    /**Internal cache of repositories*/
    private static Hashtable repositories = new Hashtable();

    private RepositoryRegistry() { }

    /**
     * Loads the registry from inputStream.
     * @param inputStream   contains the jvending-config file.
     * @throws IOException  thrown on interrupted I/O
     */
    public static synchronized void loadFromInputStream(InputStream inputStream) throws IOException {
        new RepositoryLoader().load(inputStream);
    }

    /**
     * Convenience method for loading a file off of a file system.
     * @param fileName      relative or absolute path of the file
     * @throws IOException  thrown on interrupted I/O
     */
    public static synchronized void loadFromFile(String fileName) throws IOException {
        FileInputStream fis = new FileInputStream(fileName);
        loadFromInputStream(fis);
    }


    /**
     * Convenience method for loading from a JAR or Resource.
     * @param fileName      relative or absolute path of the file
     * @throws IOException  thrown on interrupted I/O
     */
    public static synchronized void loadFromResource(String fileName, Class sourceClass) throws IOException {
        InputStream stream = sourceClass.getResourceAsStream(fileName);
        if(stream == null) throw new IOException("Could not locate resource: File Name = " + fileName);
        loadFromInputStream(stream);
    }

    /**
     * Adds a repository to the registry
     * @param name          name of the repository
     * @param repository    instance of the repository
     */
    public static synchronized void addRepository(String name, Repository repository) {
        repositories.put(name, repository);
    }

    /**
     * Finds a repository from the registry.
     * @param name  name of the repository.
     * @return  instance of the Repository
     */
    public static synchronized Repository find(String name) {
        return (Repository) repositories.get(name);
    }

    /**
     * Removes a repository from the registry
     * @param name  name of the repository
     */
    public static synchronized void removeRepository(String name) {
        repositories.remove(name);
    }

    private static class RepositoryLoader {

        /**Internal list of <code>RepositoryObject</code>s*/
        private List repositories = new ArrayList();

        /**
         * Loads the registry-config file
         * @param inputStream   inputstream containing registry-config file
         * @throws IOException
         */
        void load(InputStream inputStream) throws IOException {

            KXmlParser parser = new KXmlParser();
            try {
                parser.setInput(inputStream, null);
            } catch (XmlPullParserException e) {
               throw new IOException(e.toString());
            }
            try {
                parser.nextTag();
                parser.require(XmlPullParser.START_TAG, null, "registry-config");
                parser.nextTag();
                parser.require(XmlPullParser.START_TAG, null, "repositories");

                while(parser.nextTag() == XmlPullParser.START_TAG) {
                    parser.require(XmlPullParser.START_TAG, null, "repository");
                    repositories.add(getRepositoryObject(parser));
                }
            } catch (XmlPullParserException e) {
                throw new IOException(e.toString());
            }
            loadIntoRegistry();
        }

        /**
         * Loads all of the repositories into the registry
         * @throws IOException
         */
        private void loadIntoRegistry() throws IOException {
            for(Iterator i = repositories.iterator(); i.hasNext();) {
                RepositoryObject repositoryObject = (RepositoryObject) i.next();
                String repositoryName = repositoryObject.getRepositoryName();
                String className = repositoryObject.getRepositoryClass();
                String fileName = repositoryObject.getRepositoryConfig();
                Repository repository = loadRepository(toPath(fileName), className, repositoryObject.getInitParams());

                addRepository(repositoryName, repository);
            }
        }

        /**
         * Takes information from the registry-config file and dynamically builds a <code>Repository</code>
         * @param fileName          name of the repository's configuration file. It may be located on the file system
         *                          or within a jar.
         * @param repositoryClass   name of the repository class
         * @param initParams        <code>Hashtable</code> containing the repository's configuration parameters.
         * @return instance of repository
         * @throws IOException
         */
        private Repository loadRepository(String fileName,
                                          String repositoryClass,
                                          Hashtable initParams) throws IOException
            {
            Hashtable props = (initParams != null) ? initParams: new Hashtable();

            if(fileName == null || fileName.trim().equals("")) throw new IOException("File name must be provided.");
            if(repositoryClass == null || repositoryClass.trim().equals(""))
                throw new IOException("Repository class name must be provided: File Name = " + fileName
                        + ", Properties = " + props.toString());

            InputStream stream = null;
            Repository repository = null;
            try {
                stream = new FileInputStream(fileName);
            } catch(IOException e) {
                stream = this.getClass().getResourceAsStream(fileName);
            }
            String message = "File Name = " + fileName +
                        ", Repository Class = " + repositoryClass + ", Properties = " + props.toString();

            if(stream == null)
                throw new IOException("Unable to load config file: " + message);


            try {
                Class c = Class.forName(repositoryClass);
                repository = (Repository) c.newInstance();
                repository.load(stream, initParams);
            } catch(IOException e) {
                throw new IOException("JV-REG-001:" + e.toString() + " : " + message);
            } catch(Exception e) {
                throw new IOException("JV-REG-001:"+ e.toString() + " : " + message);
            } catch(Error e) {
                throw new IOException("JV-REG-001:" + e.toString() + " : " + message);
            }
            return repository;
        }

        /**
         * Resolves system variables within the path
         * @param fileName  name of the configuration file
         * @return  path of the file with resolved system variables. Default value is '.'
         */
         private String toPath(String fileName) {
            byte[] path = fileName.getBytes();
            int length = path.length;
            StringBuffer env = new StringBuffer();
            StringBuffer filePath = new StringBuffer();
            for(int i = 0; i < length; ) {
                if(i >=  length - 2) {
                    filePath.append((char) path[i++]);
                } else if(path[i] == 36) {
                    if(path[++i] == 123) {
                        i++;
                        while(i < length - 1 && path[i] != 125) {
                            env.append((char) path[i++]);
                        }
                        if(path[i] == 125) i++;
                    } else {
                        i--;
                        i--;
                    }
                    String pathEnv = System.getProperty(env.toString().trim(), ".");
                    filePath.append(pathEnv.toString());
                } else {
                    filePath.append((char) path[i++]);
                }
            }//end for:i
            return filePath.toString();
        }

        /**
         * Constructs a <code>RepositoryObject</code> from the registry-config file
         * @param parser
         * @return  <code>RepositoryObject</code>
         * @throws IOException
         * @throws XmlPullParserException
         */
        private RepositoryObject getRepositoryObject(KXmlParser parser) throws IOException, XmlPullParserException {
            RepositoryObject repositoryObject = new RepositoryObject();
            Hashtable initParams = new Hashtable();
            for(int i = 0; parser.nextTag() == XmlPullParser.START_TAG; i++) {
                switch(i) {
                    case 0:
                        parser.require(XmlPullParser.START_TAG, null, "repository-name");
                        repositoryObject.setRepositoryName(parser.nextText());
                        break;
                    case 1:
                        parser.require(XmlPullParser.START_TAG, null, "repository-class");
                        repositoryObject.setRepositoryClass(parser.nextText());
                    break;
                    case 2:
                        parser.require(XmlPullParser.START_TAG, null, "repository-config");
                        repositoryObject.setRepositoryConfig(parser.nextText());
                        break;
                    default:
                        parser.require(XmlPullParser.START_TAG, null, "init-param");

                        String paramName = null;
                        String paramValue = null;
                        for(int j = 0; parser.nextTag() == XmlPullParser.START_TAG; j++) {

                            switch(j) {
                                case 0:
                                    parser.require(XmlPullParser.START_TAG, null, "param-name");
                                    paramName = parser.nextText();
                                    break;
                                case 1:
                                    parser.require(XmlPullParser.START_TAG, null, "param-value");
                                    paramValue = parser.nextText();
                                    break;
                                default:
                                    throw new IOException();
                            }
                        }//end params
                        if(paramName != null && paramValue != null) initParams.put(paramName, paramValue);
                }//end all tags
                repositoryObject.setInitParams(initParams);
            }

            return repositoryObject;
        }

        /**
         * Value Object for Repository Information
         */
        private class RepositoryObject {

            /** Name of the repository */
            private String repositoryName;

           /**package and class name of the repository*/
            private String repositoryClass;

            /*Path and name of the repository config file*/
            private String repositoryConfig;

            /**Initialization parameters of the repository*/
            private Hashtable initParams;

            /**Empty Constructor*/
            RepositoryObject() { }

            /**
             * Constructor
             * @param repositoryName    name of the repository
             * @param repositoryClass   path and name of the repository config file
             * @param repositoryConfig  Path and name of the repository config fil
             */
            RepositoryObject(String repositoryName, String repositoryClass, String repositoryConfig) {
                this.repositoryName = repositoryName;
                this.repositoryClass = repositoryClass;
                this.repositoryConfig = repositoryConfig;
            }

            String getRepositoryName() {
                return repositoryName;
            }

            void setRepositoryName(String repositoryName) {
                this.repositoryName = repositoryName;
            }

            String getRepositoryClass() {
                return repositoryClass;
            }

            void setRepositoryClass(String repositoryClass) {
                this.repositoryClass = repositoryClass;
            }

            String getRepositoryConfig() {
                return repositoryConfig;
            }

            void setRepositoryConfig(String repositoryConfig) {
                this.repositoryConfig = repositoryConfig;
            }

            Hashtable getInitParams() {
                return initParams;
            }

            void setInitParams(Hashtable initParams) {
                this.initParams = initParams;
            }

            /** Classes are equal if they have the same values for class, config and name*/
            public boolean equals(Object o) {
                if (this == o) return true;
                if (!(o instanceof RepositoryObject)) return false;

                final RepositoryObject repositoryObject = (RepositoryObject) o;

                if (!repositoryClass.equals(repositoryObject.repositoryClass)) return false;
                if (!repositoryConfig.equals(repositoryObject.repositoryConfig)) return false;
                if (!repositoryName.equals(repositoryObject.repositoryName)) return false;

                return true;
            }

            /** Classes have identical hash code if they have the same values for class, config and name*/
            public int hashCode() {
                int result;
                result = repositoryName.hashCode();
                result = 29 * result + repositoryClass.hashCode();
                result = 29 * result + repositoryConfig.hashCode();
                return result;
            }

            public String toString() {
                return "Name = " + repositoryName +
                        ", Class = " + repositoryClass +
                        ", Config = " + repositoryConfig;
            }
        }
    }
}